/*
Robots.cpp
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "Robots.h"
namespace SRS10{
namespace robot
{
	void ANP::start(SRS10::robot::Robot* robot)
	{
		if(!robot)
			return;
		if(runningANP == 0)
		{
			startRot = robot->getRot();
			runningANP = 1;
			reportCollision = false;
		}
	}
	void ANP::runANP(SRS10::robot::Robot* robot, SRS10::robot::IR360* sensorANP, Overlay* hudANP)
	{
		const float precision = (float)precisionANP;
		const float maxRot = 150.0f;
		float destRot = startRot + -maxRot + maxRot*(float)navPoint/precision;
		if(runningANP == 0)
			return;

		if(!robot || !sensorANP)
			return;

		switch(runningANP)
		{
		case 1:
			for(int i=0; i<2*precisionANP+1; ++i)
				dataANP[i] = Clear;
			scanCounterANP = 0;
			rotationANP = -maxRot;
			sensorANP->set(-maxRot, 0);
			lastPointIsValidANP = false;
			++runningANP;
			if(hudANP)
				hudANP->clearAll();
			break;
		case 2:
			if(sensorANP->isRotation(-maxRot))
				++runningANP;
			break;
		case 3:
			++runningANP;
			break;
		case 4:
			if(sensorANP->isRotating())
				break;
			// get sensor read
			if(sensorANP->getLast() == 1)
			{
				float d = sensorANP->getLastDistance();
				math3D::CVector2D v;
				v.Construct(math3D::CMath3D::Deg2Rad(robot->getRot() + sensorANP->getRotation()), d*sensorANP->getRange());
				v.x += robot->getX();
				v.y += robot->getY();
				if(lastPointIsValidANP && hudANP)
				{
					hudANP->addLine(Overlay::TwoPoints(lastPointANP, Overlay::Point(v.x, v.y)));
				}
				lastPointIsValidANP = true;
				lastPointANP = Overlay::Point(v.x, v.y);
				dataANP[scanCounterANP] = Blocked;
				//hudANP.addSquare(Overlay::Point(v.x, v.y));
			}
			else if(lastPointIsValidANP)
			{
				hudANP->addLine(Overlay::TwoPoints(lastPointANP, lastPointANP));
				lastPointIsValidANP = false;
			}

			++scanCounterANP;
			if(sensorANP->isRotation(maxRot))
			{
				rotationANP = 0;
				sensorANP->set(0, 0);
				++runningANP;
				lastPointIsValidANP = false;
				runningANP = 0;

				float theta = -maxRot;
				math3D::CVector2D v;
				for(int i=0; i<2*precisionANP+1; ++i)
				{
					if(dataANP[i] == Blocked)
						continue;
					if(i >= 1 && dataANP[i-1] == Blocked)
						dataANP[i] = Warning;
					if(i < 2*precisionANP && dataANP[i+1] == Blocked)
						dataANP[i] = Warning;

					if(i >= 2 && dataANP[i-2] == Blocked)
						dataANP[i] = Warning;
					if(i < 2*precisionANP-1 && dataANP[i+2] == Blocked)
						dataANP[i] = Warning;
				}
				if(hudANP)
				{
					for(int i=0; i<2*precisionANP+1; ++i)
					{
						v.Construct(math3D::CMath3D::Deg2Rad(robot->getRot()+theta), 0.5f);
						v.x += robot->getX();
						v.y += robot->getY();

						if(dataANP[i] == Clear)
							hudANP->addSquare(Overlay::Point(v.x, v.y, 0, 1, 0));
						else if(dataANP[i] == Blocked)
							hudANP->addSquare(Overlay::Point(v.x, v.y, 1, 0, 0));
						else if(dataANP[i] == Warning)
							hudANP->addSquare(Overlay::Point(v.x, v.y, 1, 1, 0));

						theta += maxRot/precision;
					}
				}
			}
			else
			{
				rotationANP += 1.0f*maxRot/precision;
				sensorANP->set(rotationANP, 0);
			}
			break;

		case 5:
			
			if(fabs(robot->getRot() - destRot) < 1)
			{
				robot->setRotation(destRot);
				robot->runMoveCommand(1, 0);
				runningANP = 6;
			}
			else if(robot->getRot() > destRot)
				robot->runMoveCommand(4,0);
			else
				robot->runMoveCommand(3,0);
			break;

		case 6:
			if(robot->bump)
			{
				runningANP = 0;
				robot->runMoveCommand(0, 0);
				reportCollision = true;
			}
			if(math3D::CMath3D::FindDistanceSq(math3D::CVector2D(robot->getX(), robot->getY()), startPoint) > 0.25f)
			{
				runningANP = 0;
				robot->runMoveCommand(0, 0);
			}
			break;
		}
	}

	bool ANP::navToPoint(SRS10::robot::Robot* robot, int point, Overlay* hudANP)
	{
		if(runningANP != 0)
			return false;
		else if(point < 0 || point >= 2*precisionANP+1)
			return false;
		const float precision = precisionANP;
		const float maxRot = 150.0f;
		navPoint = point;
		runningANP = 5;

		startPoint.x = robot->getX();
		startPoint.y = robot->getY();

		if(hudANP)
		{
			hudANP->clearSquares();
			float theta = startRot + -maxRot + maxRot*(float)navPoint/precision;
			math3D::CVector2D v;
			v.Construct(math3D::CMath3D::Deg2Rad(theta), 0.5f);
			v.x += robot->getX();
			v.y += robot->getY();

			int i = point;
			if(dataANP[i] == Clear)
				hudANP->addSquare(Overlay::Point(v.x, v.y, 0, 1, 0));
			else if(dataANP[i] == Blocked)
				hudANP->addSquare(Overlay::Point(v.x, v.y, 1, 0, 0));
			else if(dataANP[i] == Warning)
				hudANP->addSquare(Overlay::Point(v.x, v.y, 1, 1, 0));
		}
		return true;
	}

	bool RobotManager::collision(SRS10::graphics::CollisionCircle& circle, int ID)
	{
		math3D::CVector3D bot(circle.x, circle.y, 0), v;
		
		for(int i=0; i<(int)robots.size(); ++i)
		{
			if(robots[i] && robots[i]->ID != ID)
			{
				v.x = robots[i]->x;
				v.y = robots[i]->y;
				if(sqrtf(math3D::CMath3D::FindDistanceSq2D(bot, v)) <= robots[i]->sphere+circle.size)
					return true;
			}
		}
		return false;
	}

	void IR::draw(float dx, float dy, float dr, track::Track& track, RobotManager& manager, int robotID)
	{
		math3D::CVector2D v;

		int steps=20;
		if(range > 1)
			steps = 40;
		SRS10::graphics::Circle c;
		SRS10::graphics::CollisionCircle c2;
		glColor4f(1,1,1,1);
		for(int i=0; i<steps; ++i)
		{
			v.Construct(math3D::CMath3D::Deg2Rad(rot+dr), range/steps*i);
			v.x += dx;
			v.y += dy;

			c.x1 = v.x;
			c.y1 = v.y;
			c.radius = 0.05f;

			c2.size = c.radius;
			c2.x = c.x1;
			c2.y = c.y1;
			if(track.collision(c2) || manager.collision(c2, robotID))
				break;

			//c.draw(0,0,0);
			glBegin(GL_POINTS);
				glVertex3f(c.x1, c.y1, 0);
			glEnd();
		}

		/*glBegin(GL_LINES);
			glVertex3f(dx, dy, 0);
			glVertex3f(v.x, v.y, 0);
		glEnd();*/
	}
	void LightSensor::draw(float dx, float dy, float dr, track::Track& track, RobotManager& manager, int robotID)
	{
		math3D::CVector2D v;
		v.Construct(math3D::CMath3D::Deg2Rad(dr+rot),x);
		v.x += dx;
		v.y += dy;
		graphics::Circle c;
		c.x1 = v.x;
		c.y1 = v.y;
		c.radius = 0.05f;

		glColor4f(0,0,1,1);
		c.draw(0,0,0);
		glColor4f(1,1,1,1);
	}
	int LightSensor::get(float dx, float dy, float dr, track::Track& track, RobotManager& manager, int robotID)
	{
		math3D::CVector2D v;
		v.Construct(math3D::CMath3D::Deg2Rad(dr+rot),x);
		v.x += dx;
		v.y += dy;
		if(track.lightCollision(v.x, v.y))
		{
			last = 1;
			return 1;
		}
		last = 0;
		return 0;
	}

	void LineSensor::draw(float dx, float dy, float dr, track::Track& track, RobotManager& manager, int robotID)
	{
		math3D::CVector2D v;
		v.Construct(math3D::CMath3D::Deg2Rad(dr+rot),x);
		v.x += dx;
		v.y += dy;
		graphics::Circle c;
		c.x1 = v.x;
		c.y1 = v.y;
		c.radius = 0.07f;

		glColor4f(0,1,0,1);
		c.draw(0,0,0);
		glColor4f(1,1,1,1);
	}
	int LineSensor::get(float dx, float dy, float dr, track::Track& track, RobotManager& manager, int robotID)
	{
		math3D::CVector2D v;
		v.Construct(math3D::CMath3D::Deg2Rad(dr+rot),x);
		v.x += dx;
		v.y += dy;
		SRS10::graphics::CollisionCircle c;
		c.x = v.x;
		c.y = v.y;
		c.size = 0.07f;
		if(track.getShapeList().checkNonSolidCollision(c))
		{
			cooldown = 2;
			last = 1;
			return 1;
		}
		if(cooldown > 0)
		{
			--cooldown;
			last = 1;
			return 1;
		}
		last = 0;
		return 0;
	}

	void ProximitySensor::draw(float dx, float dy, float dr, track::Track& track, RobotManager& manager, int robotID)
	{
	}
	int ProximitySensor::get(float dx, float dy, float dr, track::Track& track, RobotManager& manager, int robotID)
	{
		std::vector <Robot*>& robots = manager.getRobotList();
		Robot* r = NULL;

		int value = 0;
		int outputValue = 0;	// 0 = nothing, 1 = same_type, 2 = diff_type, 3 = both_type
		char thisType = 0;

		for(unsigned int i=0; i<robots.size(); ++i)
		{
			if(robots[i] && robots[i]->ID == robotID)
			{
				thisType = robots[i]->type;
				break;
			}
		}

		std::vector <SRS10::Entity>& entityList = SRS10::EntityManager::instance().getEntityList();
		for(unsigned int i=0; i<entityList.size(); ++i)
		{
			if(!entityList[i].alive) continue;
			math3D::CVector2D v1(entityList[i].x, entityList[i].y), v2(dx, dy);
			if(math3D::CMath3D::FindDistanceSq2D(v1, v2) > range*range)
				continue;

			float theta = atan2f(entityList[i].y- dy, entityList[i].x - dx);
			theta *= (180.0f/3.14f);
			theta -= dr;

			int t = (int)theta;
			t %= 360;
			if(t < 0)
				t = 360 + t;

			value = 0;
			switch(quad) {
				case 0:
					if(t >= 270 && t <= 360) value = 1; break;
				case 1:
					if(t >= 0 && t <= 90) value = 1; break;
				case 2:
					if(t >= 90 && t <= 180) value = 1; break;
				case 3:
					if(t >= 180 && t <= 270) value = 1; break;
				case 4:
					if((t >= 315 && t <= 360) || (t >= 0 && t <= 45)) value = 1; break;
				case 5:
					if(t >= 45 && t <= 135) value = 1; break;
				case 6:
					if(t >= 135 && t <= 225) value = 1; break;
				case 7:
					if(t >= 225 && t <= 315) value = 1; break;
			}

			if(value == 1)
			{
				outputValue = 0x04;
				break;
			}
		}

		for(unsigned int i=0; i<robots.size(); ++i)
		{
			if(!robots[i]) continue;

			r = robots[i];

			if(r->ID == robotID) continue;

			float theta = atan2f(r->getY() - dy, r->getX() - dx);
			theta *= (180.0f/3.14f);
			theta -= dr;

			int t = (int)theta;
			t %= 360;
			if(t < 0)
				t = 360 + t;

			value = 0;
			switch(quad) {
				case 0:
					if(t >= 270 && t <= 360) value = 1; break;
				case 1:
					if(t >= 0 && t <= 90) value = 1; break;
				case 2:
					if(t >= 90 && t <= 180) value = 1; break;
				case 3:
					if(t >= 180 && t <= 270) value = 1; break;
				case 4:
					if((t >= 330 && t <= 360) || (t >= 0 && t <= 30)) value = 1; break;
				case 5:
					if(t >= 30 && t <= 135) value = 1; break;
				case 6:
					if(t >= 135 && t <= 225) value = 1; break;
				case 7:
					if(t >= 225 && t <= 330) value = 1; break;
			}

			if(value == 1)
			{
				math3D::CVector2D v1(r->getX(), r->getY()), v2(dx, dy);
				if(math3D::CMath3D::FindDistanceSq2D(v1, v2) < range*range)
				{
					if(r->type == thisType)
						outputValue |= 0x01;
					else
						outputValue |= 0x02;

					if(outputValue == 3)
						break;
				}
			}
		}

		last = outputValue;
		return last;
	}

	int IR::get(float dx, float dy, float dr, track::Track& track, RobotManager& manager, int robotID)
	{
		math3D::CVector2D v;

		int steps=20;
		if(range > 1)
			steps = 40;
		SRS10::graphics::CollisionCircle c;
		for(int i=0; i<steps; ++i)
		{
			v.Construct(math3D::CMath3D::Deg2Rad(rot+dr), range/steps*i);
			v.x += dx;
			v.y += dy;

			c.x = v.x;
			c.y = v.y;
			c.size = 0.05f;

			if(track.collision(c) || manager.collision(c, robotID))
			{
				last = 1;
				return 1;
			}
		}

		last = 0;
		return 0;
	}
	
	void IR360::draw(float dx, float dy, float dr, track::Track& track, RobotManager& manager, int robotID)
	{
		math3D::CVector2D v;
		
		int steps=30;
		SRS10::graphics::Circle c;
		SRS10::graphics::CollisionCircle c2;
		if(!autoScan)
			glColor4f(1,1,0,1);
		for(int i=0; i<steps; ++i)
		{
			v.Construct(math3D::CMath3D::Deg2Rad(rot+dr), range/steps*i);
			v.x += dx;
			v.y += dy;
			
			c.x1 = v.x;
			c.y1 = v.y;
			c.radius = 0.03f;
			
			c2.size = c.radius;
			c2.x = c.x1;
			c2.y = c.y1;
			if(track.collision(c2) || manager.collision(c2, robotID))
				break;
			
			//c.draw(0,0,0);
			glBegin(GL_POINTS);
			glVertex3f(c.x1, c.y1, 0);
			glEnd();
		}
		glColor4f(1,1,1,1);
		updateRotation();
	}
	
	void IR360::set(float arg1, float arg2)
	{
		destinationRot = arg1;
	}
	
	int IR360::get(float dx, float dy, float dr, track::Track& track, RobotManager& manager, int robotID)
	{
		math3D::CVector2D v;
		
		int steps=20;
		SRS10::graphics::CollisionCircle c;
		for(int i=0; i<steps; ++i)
		{
			v.Construct(math3D::CMath3D::Deg2Rad(rot+dr), range/steps*i);
			v.x += dx;
			v.y += dy;
			
			c.x = v.x;
			c.y = v.y;
			c.size = 0.03f;
			
			if(track.collision(c) || manager.collision(c, robotID))
			{
				lastDistance = i;
				lastDistance /= steps;
				last = 1;
				return 1;
			}
		}
		
		last = 0;
		return 0;
	}

	bool Robot::loadFromFile(const std::string& filename)
	{
		cleanup();

		/*float IRRange = 1.0f;
		IRRange = 0.5f;
		sensors.push_back(new IR(0, 0, -55, IRRange));
		sensors.push_back(new IR(0, 0, 0, IRRange));
		sensors.push_back(new IR(0, 0, 55, IRRange));
		
//		sensors.push_back(new IR360(0, 0, 90, IRRange, true));
		
		sensors.push_back(new IR360(0,0,-10,IRRange*1.5f, false));
		sensors.push_back(new IR360(0,0,10, IRRange*1.5f, false));*/
//		sensors.push_back(new Timer());

		sphere = 0.125f;
		

		return shapeList.loadFromDataFile(filename);
	
//		return SRS10::graphics::loadDataFile(filename, shapeList);
	}
	
	void Robot::packEvents(wok::sockets::Packet& p, SRS10::track::Track& track, RobotManager& manager)
	{
		if(events.empty())
		{
			p.packInt(0);
			return;
		}
		std::list <Event>::iterator Iter;
		int counter = 0;
		for(Iter = events.begin(); Iter != events.end(); ++Iter)
		{
			p.packFloat(Iter->x);
			p.packFloat(Iter->y);
			p.packInt(Iter->type);
			p.packInt(Iter->lastMoveCommand);
			p.packInt(Iter->lastSpecialMoveCommand);
			p.packChar(Iter->creator);
			p.packInt(Iter->direction);

			for(unsigned int i=0; i<Iter->sensors.size(); ++i)
			{
				p.packInt(Iter->sensors[i].ID.get());
				p.packInt(Iter->sensors[i].value);
			}
			p.packChar(Iter->sensors.size());


			++counter;
		}
		
		p.packInt(counter);

		events.clear();
		return;
	}
	void Robot::packSensorValues(wok::sockets::Packet& p, SRS10::track::Track& track, RobotManager& manager)
	{
		std::vector <Sensor*>& sensors = sensorManager.getSensorVector();
		for(int i=0; i<(int)sensors.size(); ++i)
		{
			if(!sensors[i])
				continue;
			
			p.packInt(sensors[i]->get(x, y, rot, track, manager, ID));
			p.packInt(sensors[i]->getID().get());
		}
		
		p.packChar((char)sensors.size());
		
		if(bump)
			p.packChar(1);
		else
			p.packChar(0);
		
		if(robotBump)
			p.packChar(1);
		else
			p.packChar(0);

		bump = false;
		robotBump = false;
		
		p.packFloat(x);
		p.packFloat(y);
		p.packFloat(rot);
	}

	void Robot::update(SRS10::track::Track& track, RobotManager& manager, bool selected)
	{
		int FPS = SRS10::graphics::GraphicsEngine::instance().getFPS();
		float simSpeed = SRS10::graphics::GraphicsEngine::instance().getSimulationSpeed();
		float dx=0, dy=0;
		float rotChange = simSpeed*250.0f*turnSpeed/FPS;
		rotChange *= 0.2f;
		rot += rotChange;
		float forwardSpeed = simSpeed*2.0f*speed/FPS;//*(1.5f-fabs(turnSpeed));
		forwardSpeed *= 0.15f;		
		
		//if(turnSpeed < -0.5f || turnSpeed > 0.5f)
		//	forwardSpeed = 0;
		
		dx += cosf(math3D::CMath3D::Deg2Rad(rot))*forwardSpeed;
		dy += sinf(math3D::CMath3D::Deg2Rad(rot))*forwardSpeed;
		
		
		SRS10::graphics::CollisionCircle c;
		c.size = sphere;
		c.x = x+dx;
		c.y = y+dy;
		
		if(track.collision(c))
			bump = true;
	//	bump = track.collision(c);// | manager.collision(c, ID);
		robotBump = manager.collision(c, ID);
		
		SRS10::Entity e;
		if(SRS10::EntityManager::instance().checkCollision(e, x,y,sphere))
			addEvent(x, y, e.getValue(), 1);


		if(!track.collision(c) && !manager.collision(c, ID))
		{
			x += dx;
			y += dy;
		}

		if(math3D::CMath3D::FindDistanceSq2D(math3D::CVector3D(x,y,0), lastPos) <= 0.065f)
			glColor4f(1,1,0,1);
		else
			glColor4f(0,0,0,1);
		//glBegin(GL_LINES);
		//	glVertex3f(x,y,0);glVertex3f(lastPos.x, lastPos.y, 0);
		//glEnd();
		
		if(selected)
			glColor4f(0,0,1,1);
		else
			glColor4f(1,1,1,1);

		switch(type)
		{
		case 0: glColor4f(1,1,1,1); break;
		case 1: glColor4f(0,0,1,1); break;
		}

		if(selected)
			glColor4f(0,1,0,1);

		shapeList.draw(x,y,rot);
		glColor4f(1,1,1,1);

		std::vector <Sensor*>& sensors = sensorManager.getSensorVector();
		for(size_t i=0; i<sensors.size(); ++i)
		{
			if(sensors[i])
			{
				sensors[i]->get(x,y,rot, track, manager, ID);
				sensors[i]->draw(x,y,rot, track, manager, ID);
			}
		}

		if(useANP)
		{
			anp.runANP(this, &anpIR, &hudANP);
			anpIR.get(x, y, rot, track, manager, ID);
			anpIR.draw(x, y, rot, track, manager, ID);
			hudANP.draw();
		}
		
		//mainTimer.draw(0,0,0,track,manager,ID);

	}

	void Robot::addEvent(float eventX, float eventY, int eventR, char creator)
	{
		Event e;
		e.x = eventX;
		e.y = eventY;
		e.type = eventR;
		e.lastMoveCommand = lastMoveCommand;
		e.lastSpecialMoveCommand = lastSpecialMoveCommand;
		e.creator = creator;

		float theta = atan2f(e.y - y, e.x - x);
		theta *= (180.0f/3.14f);
		theta -= this->getRot();

		int t = (int)theta;
		t %= 360;
		if(t < 0)
			t = 360 + t;

		e.direction = t;

		SensorData data;

		std::vector <Sensor*>& sensors = sensorManager.getSensorVector();
		for(unsigned int i=0; i<sensors.size(); ++i)
		{
			if(!sensors[i])
				continue;

			data.ID = sensors[i]->getID();
			data.value = sensors[i]->getLast();

			e.sensors.push_back(data);
		}

		events.push_back(e);
	}
}
}