/*
SRS10Commands.cpp
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*
Main program file for the SRS10Module used by wok::Interpreter
*/

#include "SRS10Client.h"
#include "../wok/Interpreter.h"

class SRS10Module;

class EventType : public wok::interp::UserObject
{
	friend class SRS10Module;
public:
	EventType () : wok::interp::UserObject("srs10::Event") {};
	virtual ~EventType() {};

	virtual void print(std::ostream& f)
	{
		std::cout << "srs10::Event";
	}

	virtual void copy(wok::interp::UserObject* o)
	{
		if(!o || o->getType() != getType())
			return;

		e = ((EventType*)o)->e;
	}

private:
	SRS10::Event e;
};

class SRS10Module : public wok::interp::CommandLibrary
{
public:
	virtual ~SRS10Module()
	{
	};
	virtual int requiredVersion()
	{
		return 1;
	}
	virtual bool evaluate(wok::Text& text, wok::interp::Store& store, wok::interp::Object& output, wok::interp::Interpreter* interpreter, std::vector <CommandLibrary*> &library, bool globalMode)
	{
		if(text[0] == "srs10::connect")
		{
			if(text.size() <= 2)
			{
				if(client.connectToHost("127.0.0.1", 50010))
					wok::interp::ObjectHelper::newInteger(1, output);
				else
					wok::interp::ObjectHelper::newInteger(0, output);
			}
			else
			{
				// syntax is srs10::Connect IP:STRING PORT:INTEGER
				wok::interp::Object ip_obj, port_obj;
				interpreter->evaluate(text[1], store, ip_obj, library, globalMode);
				interpreter->evaluate(text[2], store, port_obj, library, globalMode);

				if(client.connectToHost(wok::interp::ObjectHelper::getString(ip_obj)->c_str(), *wok::interp::ObjectHelper::getInteger(port_obj)))
					wok::interp::ObjectHelper::newInteger(1, output);
				else
					wok::interp::ObjectHelper::newInteger(0, output);
			}
		}
		else if(text[0] == "srs10::disconnect")
		{
			client.disconnect();
		}
		else if(text[0] == "srs10::add_robot")
		{
			// syntax is srs10::add_robot NAME:STRING TYPE:integer
			wok::interp::Object name_obj, type_obj;
			interpreter->evaluate(text[1], store, name_obj, library, globalMode);
			interpreter->evaluate(text[2], store, type_obj, library, globalMode);
			int type = *wok::interp::ObjectHelper::getInteger(type_obj);
			if(client.addRobot(*wok::interp::ObjectHelper::getString(name_obj), type))
				wok::interp::ObjectHelper::newInteger(1, output);
			else
				wok::interp::ObjectHelper::newInteger(0, output);
		}
		else if(text[0] == "srs10::add_sensor")
		{
			// syntax is srs10::add_sensor ROBOT_NAME:STRING SENSOR_TYPE:STRING SENSOR_NAME:STRING[ARGS]
			wok::interp::Object robotName_obj, type_obj, name_obj;
			interpreter->evaluate(text[1], store, robotName_obj, library, globalMode);
			interpreter->evaluate(text[2], store, type_obj, library, globalMode);
			interpreter->evaluate(text[3], store, name_obj, library, globalMode);

			bool result = false;
			if(*wok::interp::ObjectHelper::getString(type_obj) == "IR")
			{
				double range, rotation;
				wok::interp::Object range_obj, rotation_obj;
				interpreter->evaluate(text[4], store, range_obj, library, globalMode);
				interpreter->evaluate(text[5], store, rotation_obj, library, globalMode);

				range = *wok::interp::ObjectHelper::getDouble(range_obj);
				rotation = *wok::interp::ObjectHelper::getDouble(rotation_obj);
				result = client.addIRSensor(*wok::interp::ObjectHelper::getString(robotName_obj), *wok::interp::ObjectHelper::getString(name_obj), (float)range, (float)rotation);
			}
			else if(*wok::interp::ObjectHelper::getString(type_obj) == "IR360")
			{
				double range, rotation;
				wok::interp::Object range_obj, rotation_obj, auto_obj;
				interpreter->evaluate(text[4], store, range_obj, library, globalMode);
				interpreter->evaluate(text[5], store, rotation_obj, library, globalMode);
				interpreter->evaluate(text[6], store, auto_obj, library, globalMode);
				bool autoScan = false;
				if(*wok::interp::ObjectHelper::getInteger(auto_obj) == 1)
					autoScan = true;

				range = *wok::interp::ObjectHelper::getDouble(range_obj);
				rotation = *wok::interp::ObjectHelper::getDouble(rotation_obj);
				result = client.addIR360Sensor(*wok::interp::ObjectHelper::getString(robotName_obj), *wok::interp::ObjectHelper::getString(name_obj), (float)range, (float)rotation, autoScan);
			}
			else if(*wok::interp::ObjectHelper::getString(type_obj) == "PROXIMITY")
			{
				double range;
				int quad;
				wok::interp::Object range_obj, quad_obj;
				interpreter->evaluate(text[4], store, range_obj, library, globalMode);
				interpreter->evaluate(text[5], store, quad_obj, library, globalMode);

				range = *wok::interp::ObjectHelper::getDouble(range_obj);
				quad = *wok::interp::ObjectHelper::getInteger(quad_obj);
				result = client.addRobotProximitySensor(*wok::interp::ObjectHelper::getString(robotName_obj), *wok::interp::ObjectHelper::getString(name_obj), (float)range, quad);
			}

			if(result)
				wok::interp::ObjectHelper::newInteger(1, output);
			else
				wok::interp::ObjectHelper::newInteger(0, output);
		}
		else if(text[0] == "srs10::poll_sensors")
		{
			// syntax is srs10::poll_sensors NAME:STRING
			wok::interp::Object name_obj;
			interpreter->evaluate(text[1], store, name_obj, library, globalMode);

			if(client.pollSensors(*wok::interp::ObjectHelper::getString(name_obj)))
				wok::interp::ObjectHelper::newInteger(1, output);
			else
				wok::interp::ObjectHelper::newInteger(0, output);
		}
		else if(text[0] == "srs10::poll_events")
		{
			// syntax is srs10::poll_events NAME:STRING
			wok::interp::Object name_obj;
			interpreter->evaluate(text[1], store, name_obj, library, globalMode);

			if(client.pollEvents(*wok::interp::ObjectHelper::getString(name_obj)))
				wok::interp::ObjectHelper::newInteger(1, output);
			else
				wok::interp::ObjectHelper::newInteger(0, output);
		}
		else if(text[0] == "srs10::create_event")
		{
			// syntax is create_event ROBOT_NAME:STRING VOLUME:double CREATOR:int reward_value:int red:double green:double blue:double
			wok::interp::Object name_obj, volume_obj, creator_obj, r_val_obj, red_obj, green_obj, blue_obj;
			interpreter->evaluate(text[1], store, name_obj, library, globalMode);
			interpreter->evaluate(text[2], store, volume_obj, library, globalMode);
			interpreter->evaluate(text[3], store, creator_obj, library, globalMode);
			interpreter->evaluate(text[4], store, r_val_obj, library, globalMode);
			interpreter->evaluate(text[5], store, red_obj, library, globalMode);
			interpreter->evaluate(text[6], store, green_obj, library, globalMode);
			interpreter->evaluate(text[7], store, blue_obj, library, globalMode);

			double volume = *wok::interp::ObjectHelper::getDouble(volume_obj);
			double red = *wok::interp::ObjectHelper::getDouble(red_obj);
			double green = *wok::interp::ObjectHelper::getDouble(green_obj);
			double blue = *wok::interp::ObjectHelper::getDouble(blue_obj);

			int creator = *wok::interp::ObjectHelper::getInteger(creator_obj);
			int r_val = *wok::interp::ObjectHelper::getInteger(r_val_obj);
			if(client.createEvent(*wok::interp::ObjectHelper::getString(name_obj), (float)volume, (char)creator, r_val, (float)red, (float)green, (float)blue))
				wok::interp::ObjectHelper::newInteger(1, output);
			else
				wok::interp::ObjectHelper::newInteger(0, output);
		}
		else if(text[0] == "srs10::get_sensor")
		{
			// syntax is srs10::get_sensor ROBOT_NAME:STRING SENSOR_NAME:STRING
			wok::interp::Object name_obj, sensor_obj;
			interpreter->evaluate(text[1], store, name_obj, library, globalMode);
			interpreter->evaluate(text[2], store, sensor_obj, library, globalMode);

			wok::interp::ObjectHelper::newInteger(client.getSensor(*wok::interp::ObjectHelper::getString(name_obj), *wok::interp::ObjectHelper::getString(sensor_obj)), output);
		}
		else if(text[0] == "srs10::move_command")
		{
			// syntax is srs10::add_robot NAME:STRING COMMAND:INTEGER
			wok::interp::Object name_obj, command_obj;
			interpreter->evaluate(text[1], store, name_obj, library, globalMode);
			interpreter->evaluate(text[2], store, command_obj, library, globalMode);

			if(client.moveCommand(*wok::interp::ObjectHelper::getString(name_obj), *wok::interp::ObjectHelper::getInteger(command_obj)))
				wok::interp::ObjectHelper::newInteger(1, output);
			else
				wok::interp::ObjectHelper::newInteger(0, output);
		}
		else if(text[0] == "srs10::events_exist")
		{
			wok::interp::Object name_obj;
			interpreter->evaluate(text[1], store, name_obj, library, globalMode);

			if(client.eventsInList(*wok::interp::ObjectHelper::getString(name_obj)))
				wok::interp::ObjectHelper::newInteger(1, output);
			else
				wok::interp::ObjectHelper::newInteger(0, output);
		}
		else if(text[0] == "srs10::get_next_event")
		{
			wok::interp::Object name_obj;
			interpreter->evaluate(text[1], store, name_obj, library, globalMode);

			EventType *t = new EventType;
			client.getNextEvent(*wok::interp::ObjectHelper::getString(name_obj), t->e);

			wok::interp::ObjectHelper::newUserObject(t, output);
		}
		else if(text[0] == "srs10::event::get")
		{
			// syntax is get EVENT FIELD
			wok::interp::Object event_obj, field_obj;
			interpreter->evaluate(text[1], store, event_obj, library, globalMode);
			interpreter->evaluate(text[2], store, field_obj, library, globalMode);

			if(wok::interp::ObjectHelper::getUserObject(event_obj)->getType() != "srs10::Event")
			{
				std::cout << "Syntax error; first argument of 'get' should be of type srs10::Event\n";
				return true;
			}

			SRS10::Event e = ((EventType*)event_obj.getValue())->e;

			if(*wok::interp::ObjectHelper::getString(field_obj) == "x")
				wok::interp::ObjectHelper::newDouble((double)e.x, output);
			else if(*wok::interp::ObjectHelper::getString(field_obj) == "y")
				wok::interp::ObjectHelper::newDouble((double)e.y, output);
			else if(*wok::interp::ObjectHelper::getString(field_obj) == "type")
				wok::interp::ObjectHelper::newInteger(e.type, output);
			else if(*wok::interp::ObjectHelper::getString(field_obj) == "creator")
				wok::interp::ObjectHelper::newInteger((int)e.creator, output);
			else if(*wok::interp::ObjectHelper::getString(field_obj) == "direction")
				wok::interp::ObjectHelper::newInteger(e.direction, output);
			else if(*wok::interp::ObjectHelper::getString(field_obj) == "direction_str")
			{
				SRS10::Client::Direction dir = client.getEventDirection(e);
				if(dir == SRS10::Client::Forward)
					wok::interp::ObjectHelper::newString("Forward", output);
				else if(dir == SRS10::Client::Right)
					wok::interp::ObjectHelper::newString("Right", output);
				else if(dir == SRS10::Client::Back)
					wok::interp::ObjectHelper::newString("Back", output);
				else if(dir == SRS10::Client::Left)
					wok::interp::ObjectHelper::newString("Left", output);
			}
			else if(*wok::interp::ObjectHelper::getString(field_obj) == "lastMoveCommand")
				wok::interp::ObjectHelper::newInteger(e.lastMoveCommand, output);
			else
				std::cout << "Syntax error; type not recognized.\n";
		}
		else
			return false;
		return true;
	}
	virtual bool newUserType(const std::string& type, const std::string& name, wok::interp::Store* store)
	{
		if(type == "srs10::Event")
		{
			EventType *t = new EventType;
			wok::interp::Object o(name, wok::interp::UserType, t);
			store->addStore(o);
		}
		return false;
	}

private:
	SRS10::Client client;
};