/*
Interpreter.cpp
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "Interpreter.h"

namespace wok
{
namespace interp
{
	void Object::add(Object& A, Object& B, Object& output)
	{
		if(output.getName() == "(wok::interpreter)temp")
			output.cleanup();
		output.setName("(wok::interpreter)temp");

		if(A.getType() == UserType)
		{
			UserObject* o = (UserObject*)A.value;
			o->add(A, B, output);
		}
		else if(B.getType() == UserType)
		{
			UserObject* o = (UserObject*)B.value;
			o->add(A, B, output);
		}
		else if(A.getType() == Integer)
		{
			if(B.getType() == Integer)
			{
				int *i = new int;
				*i = *(int*)A.value + *(int*)B.value;
				output.setType(Integer);
				output.setValue(i);
			}
			else if(B.getType() == Double)
			{
				double *v = new double;
				*v = (double) ( *(int*)A.value + *(double*)B.value );
				output.setType(Double);
				output.setValue(v);
			}
		}
		else if(A.getType() == Double)
		{
			if(B.getType() == Integer)
			{
				double *v = new double;
				*v = (double) (*(double*)A.value + *(int*)B.value);
				output.setType(Double);
				output.setValue(v);
			}
			else if(B.getType() == Double)
			{
				double *v = new double;
				*v = *(double*)A.value + *(double*)B.value;
				output.setType(Double);
				output.setValue(v);
			}
		}
	}

	void Object::subtract(Object& A, Object& B, Object& output)
	{
		if(output.getName() == "(wok::interpreter)temp")
			output.cleanup();
		output.setName("(wok::interpreter)temp");

		if(A.getType() == UserType)
		{
			UserObject* o = (UserObject*)A.value;
			o->subtract(A, B, output);
		}
		else if(B.getType() == UserType)
		{
			UserObject* o = (UserObject*)B.value;
			o->subtract(A, B, output);
		}
		else if(A.getType() == Integer)
		{
			if(B.getType() == Integer)
			{
				int *i = new int;
				*i = *(int*)A.value - *(int*)B.value;
				output.setType(Integer);
				output.setValue(i);
			}
			else if(B.getType() == Double)
			{
				double *v = new double;
				*v = (double) ( *(int*)A.value - *(double*)B.value );
				output.setType(Double);
				output.setValue(v);
			}
		}
		else if(A.getType() == Double)
		{
			if(B.getType() == Integer)
			{
				double *v = new double;
				*v = (double) (*(double*)A.value - *(int*)B.value);
				output.setType(Double);
				output.setValue(v);
			}
			else if(B.getType() == Double)
			{
				double *v = new double;
				*v = *(double*)A.value - *(double*)B.value;
				output.setType(Double);
				output.setValue(v);
			}
		}
	}

	void Object::multiply(Object& A, Object& B, Object& output)
	{
		if(output.getName() == "(wok::interpreter)temp")
			output.cleanup();
		output.setName("(wok::interpreter)temp");

		if(A.getType() == UserType)
		{
			UserObject* o = (UserObject*)A.value;
			o->multiply(A, B, output);
		}
		else if(B.getType() == UserType)
		{
			UserObject* o = (UserObject*)B.value;
			o->multiply(A, B, output);
		}
		else if(A.getType() == Integer)
		{
			if(B.getType() == Integer)
			{
				int *i = new int;
				*i = *(int*)A.value * *(int*)B.value;
				output.setType(Integer);
				output.setValue(i);
			}
			else if(B.getType() == Double)
			{
				double *v = new double;
				*v = (double) ( *(int*)A.value * *(double*)B.value );
				output.setType(Double);
				output.setValue(v);
			}
		}
		else if(A.getType() == Double)
		{
			if(B.getType() == Integer)
			{
				double *v = new double;
				*v = (double) (*(double*)A.value * *(int*)B.value);
				output.setType(Double);
				output.setValue(v);
			}
			else if(B.getType() == Double)
			{
				double *v = new double;
				*v = *(double*)A.value * *(double*)B.value;
				output.setType(Double);
				output.setValue(v);
			}
		}
	}

	void Object::divide(Object& A, Object& B, Object& output)
	{
		if(output.getName() == "(wok::interpreter)temp")
			output.cleanup();
		output.setName("(wok::interpreter)temp");

		if(A.getType() == UserType)
		{
			UserObject* o = (UserObject*)A.value;
			o->divide(A, B, output);
		}
		else if(B.getType() == UserType)
		{
			UserObject* o = (UserObject*)B.value;
			o->divide(A, B, output);
		}
		else if(A.getType() == Integer)
		{
			if(B.getType() == Integer)
			{
				int *i = new int;
				*i = *(int*)A.value / *(int*)B.value;
				output.setType(Integer);
				output.setValue(i);
			}
			else if(B.getType() == Double)
			{
				double *v = new double;
				*v = (double) ( *(int*)A.value / *(double*)B.value );
				output.setType(Double);
				output.setValue(v);
			}
		}
		else if(A.getType() == Double)
		{
			if(B.getType() == Integer)
			{
				double *v = new double;
				*v = (double) (*(double*)A.value / *(int*)B.value);
				output.setType(Double);
				output.setValue(v);
			}
			else if(B.getType() == Double)
			{
				double *v = new double;
				*v = *(double*)A.value / *(double*)B.value;
				output.setType(Double);
				output.setValue(v);
			}
		}
	}

	bool Object::compEqual(Object& other)
	{
		if(getType() == UserType)
		{
			UserObject* o = (UserObject*)this->value;
			return o->compEqual(this, &other);
		}
		else if(other.getType() == UserType)
		{
			UserObject* o = (UserObject*)other.value;
			return o->compEqual(&other, this);
		}
		else if(getType() == Integer)
		{
			if(other.getType() == Integer)
			{
				return ( *(int*)value == *(int*)other.value );
			}
			else if(other.getType() == Double)
			{
				return ( *(int*)value == *(double*)other.value );
			}
		}
		else if(getType() == Double)
		{
			if(other.getType() == Integer)
			{
				return ( *(double*)value == *(int*)other.value );
			}
			else if(other.getType() == Double)
			{
				return ( *(double*)value == *(double*)other.value );
			}
		}
		else if(getType() == String)
		{
			if(other.getType() == String)
				return ( *(std::string*)value == *(std::string*)other.value );
		}

		return false;
	}

	bool Object::compGT(Object& other)
	{
		if(getType() == UserType)
		{
			UserObject* o = (UserObject*)this->value;
			return o->compGT(this, &other);
		}
		else if(other.getType() == UserType)
		{
			UserObject* o = (UserObject*)other.value;
			return o->compGT(&other, this);
		}
		else if(getType() == Integer)
		{
			if(other.getType() == Integer)
			{
				return ( *(int*)value > *(int*)other.value );
			}
			else if(other.getType() == Double)
			{
				return ( *(int*)value > *(double*)other.value );
			}
		}
		else if(getType() == Double)
		{
			if(other.getType() == Integer)
			{
				return ( *(double*)value > *(int*)other.value );
			}
			else if(other.getType() == Double)
			{
				return ( *(double*)value > *(double*)other.value );
			}
		}

		return false;
	}

	bool Object::compLT(Object& other)
	{
		if(getType() == UserType)
		{
			UserObject* o = (UserObject*)this->value;
			return o->compLT(this, &other);
		}
		else if(other.getType() == UserType)
		{
			UserObject* o = (UserObject*)other.value;
			return o->compLT(&other, this);
		}
		else if(getType() == Integer)
		{
			if(other.getType() == Integer)
			{
				return ( *(int*)value < *(int*)other.value );
			}
			else if(other.getType() == Double)
			{
				return ( *(int*)value < *(double*)other.value );
			}
		}
		else if(getType() == Double)
		{
			if(other.getType() == Integer)
			{
				return ( *(double*)value < *(int*)other.value );
			}
			else if(other.getType() == Double)
			{
				return ( *(double*)value < *(double*)other.value );
			}
		}

		return false;
	}

	void Interpreter::evaluate(const std::string& line, Store& store, Object& output, std::vector <CommandLibrary*> &library, bool globalMode, bool copyStore)
	{
		if(output.getName() == "(wok::interpreter)temp")
			output.cleanup();
		output.clear();

		wok::Text text;
		text.addIgnore('\t');
		text.addDelimiter(' ');
		text.addDelimiter(',');
		text.addCollection(wok::Text::Pair('(', ')'));
		text.addCollection('"');

		text.parse(line);
		if(text.size() == 0)
			return;
		if(text.size() == 1)
		{
			if(text[0].size() > 0 && ((text[0][0] >= '0' && text[0][0] <= '9') || (text[0][0] == '-' && text[0].size() > 1) || (text[0][0] == '.' && text[0].size() > 1)))
			{
				// cast to integer or double:
				bool integerType = true;
				bool numberType = true;
				for(int j=0; j<(int)text[0].size(); ++j)
				{
					if(text[0][j] == '.')
					{
						if(!integerType)
						{
							numberType = false;
							break;
						}
						integerType = false;
					}
				}

				if(!numberType)
					ObjectHelper::newString(text[0], output);
				else if(integerType)
					ObjectHelper::newInteger(atoi(text[0].c_str()), output);
				else
					ObjectHelper::newDouble(atof(text[0].c_str()), output);
			}
			else
				ObjectHelper::newString(text[0], output);
			return;
		}

		Store storeCopy;
		store.copy(storeCopy);

		Store* ourStore = &storeCopy;

		if(!copyStore)
			ourStore = &store;

		if(globalMode)
			ourStore = &globalStore;

		if(text[0] == "global")
		{
			evaluate(text[1], storeCopy, output, library, true);
		}
		else if(text[0] == "begin")
		{
			for(unsigned int i=1; i<text.size(); ++i)
			{
				evaluate(text[i], storeCopy, output, library, globalMode, false);
			}
		}
		else if(text[0] == "while")
		{
			while(1)
			{
#ifdef WIN32
#ifdef ALLOW_USER_BREAK
				if(_kbhit() && getch() == 'q') throw std::string("Error: User break.");
#endif
#endif

				Object condition;
				evaluate(text[1], storeCopy, condition, library, globalMode, false);
				if(condition.getType() != Integer)
				{
					std::cout << "Syntax error; condition of while loop not integer\n";
					return;
				}
				if(*ObjectHelper::getInteger(condition) == 0)
					break;

				bool breakLoop = false;
				for(unsigned int i=2; i<text.size(); ++i)
				{
					evaluate(text[i], storeCopy, output, library, globalMode, false);
					if(output.getValue() && output.getType() == String)
					{
						if(*ObjectHelper::getString(output) == "break")
						{
							breakLoop = true;
							break;
						}
						else if(*ObjectHelper::getString(output) == "continue")
						{
							break;
						}
					}
				}
				if(breakLoop)
					break;
			}
		}
		else if(text[0] == "if")
		{
			Object test;
			evaluate(text[1], storeCopy, test, library, globalMode);
			if(*ObjectHelper::getInteger(test) != 0)
				evaluate(text[2], storeCopy, output, library, globalMode);
			else
				evaluate(text[3], storeCopy, output, library, globalMode);
		}
		else if(text[0] == "case")
		{
			Object test;
			evaluate(text[1], storeCopy, test, library, globalMode);
			for(int i=2; i<(int)text.size(); i+=2)
			{
				Object test2;
				evaluate(text[i], storeCopy, test2, library, globalMode);
				if(test2.getType() == String && i == (int)text.size()-1)
					break;
				if(i+1 >= (int)text.size())
					break;
				if((test2.getType() == String && *ObjectHelper::getString(test2) == "default") || test.compEqual(test2))
				{
					evaluate(text[i+1], storeCopy, output, library, globalMode);
					break;
				}
			}
		}
		else if(text[0] == "cond")
		{
			for(int i=1; i<(int)text.size(); i+=2)
			{
				Object test;
				evaluate(text[i], storeCopy, test, library, globalMode);
				if(test.getType() == String && i == (int)text.size()-1)
					break;
				if(i+1 >= (int)text.size())
					break;
				if(*ObjectHelper::getInteger(test) != 0)
				{
					evaluate(text[i+1], storeCopy, output, library, globalMode);
					break;
				}
			}
		}
		else if(text[0] == "nop" || text[0] == "NOP")
		{
		}
		else if(text[0] == "path")
		{
			Object p;
			evaluate(text[1], storeCopy, p, library, globalMode);
			addToPath(*ObjectHelper::getString(p));
		}
		else if(text[0] == "file")
		{
			Object filename;
			std::ifstream f;
			f.open(text[1].c_str());
			if(!f.is_open())
			{
				for(int i=0; i<(int)path.size(); ++i)
				{
					std::stringstream str;
					str << path[i] << "/" << text[1].c_str();
					f.open(str.str().c_str());
					if(f.is_open()) break;
				}

				if(!f.is_open())
				{
					std::cout << "Cannot open file '" << text[1] << "'\n";
					return;
				}
			}

			int maxBuffer = 2048;
			char* buffer = new char[maxBuffer+2];
			std::stringstream str;
			while(!f.eof())
			{
				char c = f.get();
				if(c == 10 || c == 13)
					str << " ";
				else
					str << c;
			}

			std::string newLine(str.str());
			preProcess(newLine);

			evaluate(newLine, storeCopy, output, library, globalMode, false);

			delete[] buffer;
		}
		else if(text[0] == "load_module")
		{
			Object val;
			evaluate(text[1], storeCopy, val, library, globalMode);
			for(int i=0; i<(int)modules.size(); ++i)
			{
				if(modules[i].name == *ObjectHelper::getString(val))
				{
					std::cout << "Module alreadly loaded.\n";
					return;
				}
			}
#ifdef ALLOW_MODULES
#ifdef WIN32
			HMODULE h = LoadLibrary(ObjectHelper::getString(val)->c_str());
			if(h == NULL)
			{
				for(int i=0; i<(int)path.size(); ++i)
				{
					std::stringstream str;
					str << path[i] << "/" << *ObjectHelper::getString(val);
					h = LoadLibrary(str.str().c_str());
					if(h != NULL) break;
				}

				if(h == NULL)
				{
					std::cout << "Cannot load module '" << *ObjectHelper::getString(val) << "'\n";
					return;
				}
			}
			moduleCallBack m = (moduleCallBack)GetProcAddress(h, "LoadCommandLibrary");
			if(!m)
			{
				std::cout << "Error " << GetLastError() << "\n";
				FreeLibrary(h);
				std::cout << "Cannot run command 'LoadCommandLibrary'\n";
				return;
			}
			CommandLibrary* lib = m();
			if(!lib)
			{
				FreeLibrary(h);
				std::cout << "'LoadCommandLibrary()' returned NULL.\n";
				return;
			}
			moduleHandles.push_back(h);
			modules.push_back(Module(*ObjectHelper::getString(val), lib));
#endif
#endif
		}
		else if(text[0] == "print")
		{
			Object val;
			evaluate(text[1], storeCopy, val, library, globalMode);
			val.print(std::cout, true);
			std::cout << "\n";
		}
		else if(text[0] == "print_new_line")
		{
			std::cout << "\n";
		}
		else if(text[0] == "print_one_line")
		{
			Object val;
			evaluate(text[1], storeCopy, val, library, globalMode);
			val.print(std::cout, false);
		}
		else if(text[0] == "integer")
		{
			ObjectHelper::newInteger(atoi(text[1].c_str()), output);
		}
		else if(text[0] == "double")
		{
			ObjectHelper::newDouble(atof(text[1].c_str()), output);
		}
		else if(text[0] == "string")
		{
			ObjectHelper::newString(text[1], output);
		}
		else if(text[0] == "+")
		{
			Object A, B;
			evaluate(text[1], storeCopy, A, library, globalMode);
			evaluate(text[2], storeCopy, B, library, globalMode);

			output.add(A, B, output);
		}
		else if(text[0] == "-")
		{
			Object A, B;
			evaluate(text[1], storeCopy, A, library, globalMode);
			evaluate(text[2], storeCopy, B, library, globalMode);

			output.subtract(A, B, output);
		}
		else if(text[0] == "*")
		{
			Object A, B;
			evaluate(text[1], storeCopy, A, library, globalMode);
			evaluate(text[2], storeCopy, B, library, globalMode);

			output.multiply(A, B, output);
		}
		else if(text[0] == "/")
		{
			Object A, B;
			evaluate(text[1], storeCopy, A, library, globalMode);
			evaluate(text[2], storeCopy, B, library, globalMode);

			output.divide(A, B, output);
		}
		else if(text[0] == "comp.eq" || text[0] == "==")
		{
			Object A, B;
			evaluate(text[1], storeCopy, A, library, globalMode);
			evaluate(text[2], storeCopy, B, library, globalMode);

			bool result = A.compEqual(B);
			if(result)
				ObjectHelper::newInteger(1, output);
			else
				ObjectHelper::newInteger(0, output);
		}
		else if(text[0] == "comp.lt" || text[0] == "<")
		{
			Object A, B;
			evaluate(text[1], storeCopy, A, library, globalMode);
			evaluate(text[2], storeCopy, B, library, globalMode);

			bool result = A.compLT(B);
			if(result)
				ObjectHelper::newInteger(1, output);
			else
				ObjectHelper::newInteger(0, output);
		}
		else if(text[0] == "comp.gt" || text[0] == ">")
		{
			Object A, B;
			evaluate(text[1], storeCopy, A, library, globalMode);
			evaluate(text[2], storeCopy, B, library, globalMode);

			bool result = A.compGT(B);
			if(result)
				ObjectHelper::newInteger(1, output);
			else
				ObjectHelper::newInteger(0, output);
		}
		else if(text[0] == "not")
		{
			Object arg;
			evaluate(text[1], storeCopy, arg, library, globalMode);
			if(arg.getType() != Integer)
			{
				std::cout << "Syntax error '" << line << "'; 'and' argument not integer.\n";
				return;
			}
			if(*ObjectHelper::getInteger(arg) == 0)
				ObjectHelper::newInteger(1, output);
			else
				ObjectHelper::newInteger(0, output);
		}
		else if(text[0] == "and")
		{
			Object arg;
			int result = 1;
			for(int i=1; i<(int)text.size(); ++i)
			{
				evaluate(text[i], storeCopy, arg, library, globalMode);
				if(arg.getType() == String && i == (int)text.size()-1)
					break;
				if(arg.getType() != Integer)
				{
					std::cout << "Syntax error '" << line << "'; 'and' argument not integer.\n";
					return;
				}
				if(i == 1)
					result = *ObjectHelper::getInteger(arg);
				else
					result &= *ObjectHelper::getInteger(arg);
			}

			ObjectHelper::newInteger(result, output);
		}
		else if(text[0] == "or")
		{
			Object arg;
			int result = 0;
			for(int i=1; i<(int)text.size(); ++i)
			{
				evaluate(text[i], storeCopy, arg, library, globalMode);
				if(arg.getType() == String && i == (int)text.size()-1)
					break;
				if(arg.getType() != Integer)
				{
					std::cout << "Syntax error '" << line << "'; 'and' argument not integer.\n";
					return;
				}
				if(i == 1)
					result = *ObjectHelper::getInteger(arg);
				else
					result |= *ObjectHelper::getInteger(arg);
			}

			ObjectHelper::newInteger(result, output);
		}
		else if(text[0] == "get")
		{
			Object name;
			evaluate(text[1], storeCopy, name, library, globalMode);
			ourStore->get(*ObjectHelper::getString(name), output);
		}
		else if(text[0] == "set")
		{
			Object name, value;
			evaluate(text[1], storeCopy, name, library, globalMode);
			evaluate(text[2], storeCopy, value, library, globalMode);

			ourStore->set(*ObjectHelper::getString(name), value);
		}
		else if(text[0] == "user.input")
		{
			Object type_obj;
			evaluate(text[1], storeCopy, type_obj, library, globalMode);

			if(*ObjectHelper::getString(type_obj) == "integer")
			{
				char buffer[101];
				std::cout << "\nPlease enter an integer: ";
				std::cin.getline(buffer, 100);
				ObjectHelper::newInteger(atoi(buffer), output);
			}
			else if(*ObjectHelper::getString(type_obj) == "char")
			{
				std::cout << "\nPlease press a key: ";
#ifdef WIN32
				ObjectHelper::newInteger(getche(), output);
#endif
				std::cout << "\n";
			}
			else if(*ObjectHelper::getString(type_obj) == "string")
			{
				char buffer[101];
				std::cout << "\nPlease enter a string: ";
				std::cin.getline(buffer, 100);
				std::string str(buffer);
				ObjectHelper::newString(str, output);
			}
			else if(*ObjectHelper::getString(type_obj) == "kbhit")
			{
#ifdef WIN32
				if(_kbhit())
				{
					char c = getch();
					ObjectHelper::newInteger((int)c, output);
				}
				else
					ObjectHelper::newInteger(0, output);
#else
				ObjectHelper::newInteger(0, output);
#endif
			}
		}
		else if(text[0] == "index")
		{
			Object name, index;
			evaluate(text[1], storeCopy, name, library, globalMode);
			std::stringstream arrayName;
			arrayName << *ObjectHelper::getString(name);
			for(int i=2; i<(int)text.size(); ++i)
			{
				evaluate(text[i], storeCopy, index, library, globalMode);

				if(name.getType() != String)
				{
					std::cout << "Syntax error '" << line << "'\n";
					return;
				}

			
				if(index.getType() == Integer)
					arrayName << "." << *ObjectHelper::getInteger(index);
				else if(index.getType() == String)
				{
					if(ObjectHelper::getString(index)->size() > 0 && (*ObjectHelper::getString(index))[0] != ' ')
						arrayName << "." << *ObjectHelper::getString(index);
				}
				else
				{
					std::cout << "Syntax error '" << line << "'; index not integer or string.\n";
					return;
				}
			}
			ObjectHelper::newString(arrayName.str(), output);
		}
		else if(text[0] == "new")
		{
			Object type, name;
			evaluate(text[1], storeCopy, type, library, globalMode);
			evaluate(text[2], storeCopy, name, library, globalMode);

			if(*ObjectHelper::getString(name) == "(wok::interpreter)temp")
			{
				std::cout << "Illegal name of variable; on line '" << line << "'\n";
				return;
			}

			std::string type_str(*ObjectHelper::getString(type));
			int amount = 1;
			bool arrayMode = false;
			if(*ObjectHelper::getString(type) == "array")
			{
				// (new array NAME TYPE DIM)
				Object type_obj, dim_obj;
				evaluate(text[3], storeCopy, type_obj, library, globalMode);
				evaluate(text[4], storeCopy, dim_obj, library, globalMode);

				amount = *ObjectHelper::getInteger(dim_obj);
				type_str = *ObjectHelper::getString(type_obj);

				int *newInt = new int;
				*newInt = amount;
				std::stringstream str;
				str << *ObjectHelper::getString(name) << ".size";
				Object o(str.str(), Integer, newInt);
				ourStore->addStore(o);
				arrayMode = true;
			}

			for(int i=0; i<amount; ++i)
			{
				std::stringstream name_str;
				name_str << *ObjectHelper::getString(name);
				if(arrayMode)
					name_str << "." << i;
				if(type_str == "integer")
				{
					int *newInt = new int;
					*newInt = 0;

					Object o(name_str.str(), Integer, newInt);
					ourStore->addStore(o);
				}
				else if(type_str == "double")
				{
					double *newDouble = new double;
					*newDouble = 0;

					Object o(name_str.str(), Double, newDouble);
					ourStore->addStore(o);
				}
				else if(!library.empty())
				{
					for(int i=0; i<(int)library.size(); ++i)
					{
						if(!library[i]) continue;
						if(library[i]->newUserType(type_str, name_str.str(), ourStore))
							break;
					}
				}
			}
		}
		else
		{
			bool commandFound = false;
			for(int i=0; i<(int)library.size(); ++i)
			{
				if(!library[i]) continue;
				if(library[i]->requiredVersion() > InterpreterVersion) continue;
				if(library[i]->evaluate(text, storeCopy, output, this, library, globalMode))
				{
					commandFound = true;
					break;
				}
			}
			if(!commandFound)
			{
				for(int i=0; i<(int)modules.size(); ++i)
				{
					if(!modules[i].library) continue;
					if(modules[i].library->requiredVersion() > InterpreterVersion) continue;
					if(modules[i].library->evaluate(text, storeCopy, output, this, library, globalMode))
					{
						commandFound = true;
						break;
					}
				}
			}
			if(!commandFound)
			{
				std::cout << "Syntax error '" << line << "'\n";
			}
		}
	}
}
}